Class w004875a_EuoplocephalusTutus_Pawn Extends GamePawn;

//If false, the idle animation is used and all movement animations are skipped.
Var EditConst Bool AnimationMoving;

//References to the movement state crossfaders, populated in PostBeginPlay and used in Tick.
Var EditConst AnimNodeCrossfader AnimationSneakWalkBlend;
Var EditConst AnimNodeCrossfader AnimationWalkRunBlend;

//If true, the walk/run left animations are mirrorred.
Var EditConst Bool AnimationTurningRight;

//References to the turn crossfaders, populated in PostBeginPlay and used in Tick.
Var EditConst AnimNodeCrossfader AnimationWalkTurningBlend;
Var EditConst AnimNodeCrossfader AnimationRunTurningBlend;

//The names of the four feet of the dinosaur, used for spawning particle effects and sounds.
Var Const Array<Name> AnimationFootBoneNames;

//A list of sounds to play, randomly chosen, on foot steps.
//Will favour sounds towards the end of the list at high speed.
Var Const Array<SoundCue> AnimationFootSounds;


Var EditConst Float Speed;
Var EditConst Float Turn;


//An offset applied to the camera target height.
Var EditConst Float CameraBouncePosition;

//The velocity to be applied to the camera target height.
Var EditConst Float CameraBounceVelocity;

//For every unit that the camera is misaligned, the amount of velocity that should be applied to try and correct it.
Var Const Float CameraBounceCentering;

//Zero will cause the camera to oscillate freely, while increasing values will damp the oscillation.  Infinity prevents camera motion.
Var Const Float CameraBounceDamping;

//The amount of force to apply to the camera at top speed.
Var Const Float CameraBounceImpact;

//A list of camera positions which can be switched between using the jump key.
//X axis - Distance.
//Y axis - Height.
//Z axis - Yaw angle, in radians.  0 is behind, PI is ahead.
Var Const Array<Vector> CameraLocations;

Var EditConst Int CameraLocationSlot;


Var EditConst ParticleSystemComponent BreathEmitter;

//Finds references to the animation nodes directly controlled by this script.
//Also attaches the breath emitter to the head bone.
Event PostBeginPlay()
{
	
	AnimationSneakWalkBlend = AnimNodeCrossfader( Mesh.FindAnimNode( 'SneakWalkBlend' ) );
	AnimationWalkRunBlend = AnimNodeCrossfader( Mesh.FindAnimNode( 'WalkRunBlend' ) );
	
	AnimationWalkTurningBlend = AnimNodeCrossfader( Mesh.FindAnimNode( 'WalkTurningBlend' ) );
	AnimationRunTurningBlend = AnimNodeCrossfader( Mesh.FindAnimNode( 'RunTurningBlend' ) );

	//Sanity check.
	If
	(
		( AnimationSneakWalkBlend == None ) ||
		( AnimationWalkRunBlend == None ) ||
		( AnimationWalkTurningBlend == None ) ||
		( AnimationRunTurningBlend == None )
	)
		`Warn( Self$" failed to find animation nodes during PostBeginPlay.  Please ensure that the correct AnimTree is in use and if one is selected in an extended PostBeginPlay, that it is set before calling the superclass's implementation." );

	Mesh.AttachComponentToSocket( BreathEmitter, 'BreathSocket' );

	Super.PostBeginPlay();

}

//Updates the animation controls to match the state of the dinosaur.
Simulated Event Tick( Float FrameDelta )
{

	Local Float SneakWalkBlend, WalkRunBlend;

	Local Vector XAxis, YAxis, ZAxis;
	

	//Get vectors pointing forward, sidewards and upwards relative to the dinosaur.
	GetAxes( Rotation, XAxis, YAxis, ZAxis );


	//Build values
	Speed = ( Velocity Dot XAxis ) / GroundSpeed;

	SneakWalkBlend = FClamp( Abs( Speed ) * 2.0f, 0.0f, 1.0f );
	WalkRunBlend = FClamp( ( Speed * 2.0f ) - 1.0f, 0.0f, 1.0f );

	
	//A value, from -1.0f to 1.0f, representing how much the velocity of the dinosaur is going sidewards.
	Turn = Normal( Velocity ) Dot YAxis;


	//Apply the movement animation.
	AnimationMoving = Abs( Speed ) > 0.01f;
	AnimationSneakWalkBlend.SetBlendTarget( SneakWalkBlend, 0.0f );
	AnimationWalkRunBlend.SetBlendTarget( WalkRunBlend, 0.0f );

	//Speed up animations as we go faster; they run at the same ground speed despite being different stances for easier synchronization.
	Mesh.GlobalAnimRateScale = Lerp( 1.0f, 3.0f, FClamp( Abs( Speed ), 0.0f, 1.0f ) );

	
	//Apply the turning animation.
	//It is not necessary to negate these at low speed, as the sneak/idle animations do not have turn animations.
	AnimationTurningRight = Turn < 0.0f;
	AnimationWalkTurningBlend.SetBlendTarget( FClamp( Abs( Turn ), 0.0f, 1.0f ) ** 0.7f, 0.0f );
	AnimationRunTurningBlend.SetBlendTarget( FClamp( Abs( Turn ), 0.0f, 1.0f ) ** 0.7f, 0.0f );


	Super.Tick( FrameDelta );

}

//Calculates the location and rotation of the camera.
//Also updates the camera bounce.
//Correct integration algorithm is based on the findings of:
//http://www.niksula.hut.fi/~hkankaan/Homepages/gravity.html
//(Doing Gravity Right :: Hannu's Plaza)
Simulated Function Bool CalcCamera( Float FrameDelta, Out Vector CameraLocation, Out Rotator CameraRotation, Out Float CameraFOV )
{

	Local Float CameraAcceleration;
	Local Float Damping;

	Local Vector XAxis, YAxis, ZAxis;


	GetAxes( Rotation, XAxis, YAxis, ZAxis );


	Damping = 1.0f + ( FrameDelta * CameraBounceDamping * 0.25f );

	CameraAcceleration = CameraBouncePosition * -CameraBounceCentering * FrameDelta;

	CameraLocation = 
		Location -

		(

		YAxis * Sin( CameraLocations[ CameraLocationSlot ].Z ) + 
		XAxis * Cos( CameraLocations[ CameraLocationSlot ].Z )

		) * CameraLocations[ CameraLocationSlot ].X +

		Vect( 0.0f, 0.0f, 1.0f ) * 
		CameraLocations[ CameraLocationSlot ].Y;

	CameraBounceVelocity /= Damping;
	CameraBounceVelocity += CameraAcceleration;
	CameraBounceVelocity /= Damping;
	CameraBouncePosition += CameraBounceVelocity * FrameDelta;
	CameraBounceVelocity /= Damping;
	CameraBounceVelocity += CameraAcceleration;
	CameraBounceVelocity /= Damping;

	CameraRotation = Rotator
	(

		Location + 
		( Vect( 0.0f, 0.0f, 1.0f ) * CameraBouncePosition ) + 
		Vect( 0.0f, 0.0f, 192.0f ) -
		CameraLocation

	);

	CameraLocation -= Vect( 0.0f, 0.0f, 35.0f );

	CameraFOV = 70.0f;

	

	Return True;

}

//Called by AnimNotify_FootStep events in the AnimSet.
Simulated Event PlayFootStepSound( Int FootDown )
{

	WorldInfo.MyEmitterPool.SpawnEmitter
	(

		ParticleSystem'JamesWild_EuoplocephalusTutus.Effects.Dust_P', 
		Mesh.GetBoneLocation( AnimationFootBoneNames[ FootDown % 4 ] )

        );

	If( Speed > 0.5f )
		WorldInfo.MyEmitterPool.SpawnEmitter
		(

			ParticleSystem'JamesWild_EuoplocephalusTutus.Effects.Cloud_P',
			Mesh.GetBoneLocation( AnimationFootBoneNames[ FootDown % 4 ] )

		);

	PlaySound
	(

		AnimationFootSounds[ Clamp( Int( AnimationFootSounds.Length * Speed ), 0, AnimationFootSounds.Length - 4 ) + Rand( 4 ) ],
		, , , Mesh.GetBoneLocation( AnimationFootBoneNames[ FootDown % 4 ] )

	);
		

	CameraBounceVelocity += CameraBounceImpact * Speed;

}

Event BreatheBegin()
{

	BreathEmitter.ActivateSystem();

}

Event BreatheEnd()
{

	BreathEmitter.DeactivateSystem();

}

Function Bool DoJump( Bool Updating )
{

	CameraLocationSlot = ( CameraLocationSlot + 1 ) % CameraLocations.Length;

	Return False;

}

//Puts the skeletal mesh into a ragdolling state.
//First, the physics mode must be set to PHYS_RigidBody and the bones unlocked from the animation system.
//By default, the rigid bodies cannot collide with most objects, only on their own channel.  (Untitled3)
//This must be changed so that the physics asset collides with the world and objects inside it.
//Finally, a blend control exists for smoothly transitioning between an animation in progress and the physics engine, which is set fully to the physics engine.
Function Bool Died( Controller Killer, Class<DamageType> DamageType, Vector HitLocation )
{

	SetPhysics( PHYS_RigidBody );
	Mesh.PhysicsAssetInstance.SetAllBodiesFixed( False );

	Mesh.SetRBChannel( RBCC_Pawn );
	Mesh.SetRBCollidesWithChannel( RBCC_Default, True );
	Mesh.SetRBCollidesWithChannel( RBCC_Pawn, True );
	Mesh.SetRBCollidesWithChannel( RBCC_Vehicle, True );
	Mesh.SetRBCollidesWithChannel( RBCC_BlockingVolume, True );
	Mesh.SetRBCollidesWithChannel( RBCC_Untitled3, False );

	Mesh.PhysicsWeight = 1.0f;

	BreathEmitter.DeactivateSystem();

	Return Super.Died( Killer, DamageType, HitLocation );

}

//Disable the code that handles the breath emitter and footsteps.
State Dying
{

	Ignores BreatheBegin, BreatheEnd, PlayFootStepSound;

}

DefaultProperties
{


	//Add a LightEnvironment.
	//This is the object which calculates the lighting strength/colour/direction/etc. of our Pawn.
 	Begin Object Class=DynamicLightEnvironmentComponent Name=MyLightEnvironment

		//Use a spherical harmonics approximation of the lights in the environment.
		bSynthesizeSHLight = True;

		//If this is true, the renderer only does a shadow test at the origin, ignoring shadow maps.
		bUseBooleanEnvironmentShadowing = False;

		bIsCharacterLightEnvironment = True;

 	End Object

	//Add the LightEnvironment to the Components array.
 	Components.Add( MyLightEnvironment );


	//Add the skeletal mesh the player sees.
 	Begin Object Class=SkeletalMeshComponent Name=PawnMesh
		
		//Use the above LightEnvironment.
		LightEnvironment = MyLightEnvironment;

		//Select our assets.
		SkeletalMesh = SkeletalMesh'JamesWild_EuoplocephalusTutus.Dinosaur.Final_Mesh';
		AnimSets = ( AnimSet'JamesWild_EuoplocephalusTutus.Animation.Final_Animset' );
		AnimTreeTemplate = AnimTree'JamesWild_EuoplocephalusTutus.Animation.Final_Animtree';
		PhysicsAsset = PhysicsAsset'JamesWild_EuoplocephalusTutus.Animation.Final_Mesh_Physics';

		//Move the model down as the CylinderComponent's origin is in the middle of the feet.
		Translation = ( X = 0.0f, Y = 0.0f, Z = -76.0f );

		//Turn the model 180 degrees about the Z axis.
		Rotation = ( Pitch = 0, Yaw = 32768, Roll = 0 );

		//Required to enable ragdolling on death.
		//See the implementation of Died.
		BlockRigidBody = True;
		RBChannel = RBCC_Untitled3;
		RBCollideWithChannels = ( Untitled3 = True );
		bHasPhysicsAssetInstance = True;

 	End Object

	//Add the SkeletalMeshComponent to the Components array and set the base class reference.
 	Components.Add( PawnMesh );
 	Mesh = PawnMesh;


	Begin Object Class=ParticleSystemComponent Name=BreathEmitter

		bAutoActivate = False;
		Template = ParticleSystem'JamesWild_EuoplocephalusTutus.Effects.Breath_P';

	End Object
	
	Components.Add( BreathEmitter );
	BreathEmitter = BreathEmitter;


	Begin Object Name=CollisionCylinder

		CollisionRadius = 34.0f;

		//NOTE: Change PawnMesh's translation if this is altered.
		CollisionHeight = 78.0f;

	End Object

	AnimationFootBoneNames = ( "CATRigLFLegAnkle", "CATRigRFLegAnkle", "CATRigLLegAnkle", "CATRigRLegAnkle" );
	AnimationFootSounds = ( SoundCue'A_Character_Footsteps.FootSteps.A_Character_Footstep_FoliageCue', SoundCue'A_Character_Footsteps.FootSteps.A_Character_Footstep_FoliageJumpCue', SoundCue'A_Character_Footsteps.FootSteps.A_Character_Footstep_StoneCue', SoundCue'A_Character_Footsteps.FootSteps.A_Character_Footstep_StoneJumpCue', SoundCue'A_Character_Footsteps.FootSteps.A_Character_Footstep_FleshCue', SoundCue'A_Character_Footsteps.FootSteps.A_Character_Footstep_FleshJumpCue', SoundCue'A_Character_Footsteps.FootSteps.A_Character_Footstep_MudLandCue', SoundCue'A_Character_Footsteps.FootSteps.A_Character_Footstep_MudJumpCue' );

	//Motion properties in the base class.  This increases the top speed of the pawn, but reduces the acceleration.  Should take 3 seconds to reach top speed, which is better for a heavy dinosaur.
	GroundSpeed = 768.0f;
	AccelRate = 256.0f;

	CameraBounceCentering = 128.0f;
	CameraBounceDamping = 42.0f;
	CameraBounceImpact = 192.0f;

	//NOTE: I'd typically put this opening bracket on the next line in my coding style, but cannot here due to a parser bug.
	CameraLocations = {(

		( X = 1280.0f, Y = 768.0f, Z = 0.0f ),
		( X = 768.0f, Y = 512.0f, Z = 3.0f ),
		( X = 768.0f, Y = 0.0f, Z = 2.0f ),
		( X = 4.0f, Y = 1024.0f, Z = 0.0f ),
		( X = 512.0f, Y = 256.0f, Z = 5.0f ),
		( X = 512.0f, Y = 0.0f, Z = 3.5f )

	)};

}